# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=too-few-public-methods, protected-access, too-many-nested-blocks, too-many-branches

from urllib.parse import parse_qs, urljoin, urlparse, quote
import json

from azure.core.exceptions import ClientAuthenticationError, ResourceExistsError, ResourceNotFoundError, \
    HttpResponseError
from azure.cli.core.azclierror import InvalidArgumentValueError

from ._arg_browser import AAZArgBrowser
from ._base import AAZUndefined, AAZBaseValue, AAZBaseType, has_value
from ._content_builder import AAZContentBuilder
from ._field_type import AAZSimpleType, AAZObjectType, AAZBaseDictType, AAZListType
from ._field_value import AAZList, AAZObject, AAZBaseDictValue
from .exceptions import AAZInvalidValueError


class AAZOperation:

    def __init__(self, ctx):
        self.ctx = ctx


class AAZHttpOperation(AAZOperation):
    """ Http Operation
    """

    CLIENT_TYPE = None  # http client registered, its value should be in the keys of aaz._client.registered_clients

    def __init__(self, ctx):
        super().__init__(ctx)
        self.client = ctx.get_http_client(self.CLIENT_TYPE)
        # common http errors by status code
        self.error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
        }

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()

    @staticmethod
    def serialize_url_param(name, value, required=True, skip_quote=False, **kwargs):  # pylint: disable=unused-argument
        if isinstance(value, AAZBaseValue):
            value = value.to_serialized_data()

        if value == AAZUndefined or value == None:  # noqa: E711, pylint: disable=singleton-comparison
            if required:
                raise ValueError(f"url parameter {name} is required.")
            return {}  # return empty dict

        if isinstance(value, (list, dict)):
            raise NotImplementedError(f"not support type {type(value)} for url parameter")

        if isinstance(value, bool):
            value = json.dumps(value)

        if skip_quote is True:
            value = str(value)
        else:
            value = quote(str(value), safe='')

        return {name: value}

    @staticmethod
    def serialize_query_param(name, value, required=False, skip_quote=False, **kwargs):
        if isinstance(value, AAZBaseValue):
            value = value.to_serialized_data()

        if value == AAZUndefined:
            if required:
                raise ValueError(f"query parameter {name} is required.")
            return {}

        def process_element(e):
            if isinstance(e, (list, dict)):
                raise NotImplementedError(f"Not support {type(e)} type element")

            if isinstance(e, bool):
                e = json.dumps(e)
            elif e is None:
                e = ""

            if skip_quote is True:
                e = str(e)
            else:
                e = quote(str(e), safe='')
            return e

        if isinstance(value, list):
            value = [process_element(v)
                     for v in value if v != AAZUndefined]
            # Determines the format of the array. Possible formats are:
            #   csv - comma separated values 'foo,bar'
            #   ssv - space separated values 'foo bar'
            #   tsv - tab separated values 'foo\tbar'
            #   pipes - pipe separated values 'foo|bar'
            # default is csv format
            div = kwargs.get('div', ',')
            if div:
                value = div.join(value)
            value = str(value)
        else:
            # Not a list
            value = process_element(value)
        return {name: value}

    @staticmethod
    def serialize_header_param(name, value, required=False, **kwargs):  # pylint: disable=unused-argument
        if isinstance(value, AAZBaseValue):
            value = value.to_serialized_data()

        if value == AAZUndefined:
            if required:
                raise ValueError(f"query parameter {name} is required.")
            return {}

        def process_element(e):
            if isinstance(e, (list, dict)):
                raise NotImplementedError(f"Not support {type(e)} type element")

            if isinstance(e, bool):
                e = json.dumps(e)
            elif e is None:
                e = ""
            return e

        if isinstance(value, list):
            value = [process_element(v) for v in value if v != AAZUndefined]
        else:
            value = process_element(value)
        value = str(value)
        return {name: value}

    @staticmethod
    def serialize_content(value, required=False):

        def processor(schema, result):
            if schema._flags.get('read_only', False):
                # ignore read_only fields when serialize content
                return AAZUndefined
            if result == AAZUndefined or result is None:
                return result

            if isinstance(schema, AAZObjectType):
                assert isinstance(result, dict)
                for _schema in [schema, schema.get_discriminator(result)]:
                    if not _schema:
                        continue
                    for _field_name, _field_schema in _schema._fields.items():
                        # verify required and not read only property
                        _name = _field_schema._serialized_name or _field_name  # prefer using serialized name first
                        if _name in result:
                            continue
                        if _field_schema._flags.get('read_only', False):
                            continue
                        if not _field_schema._flags.get('required', False):
                            continue

                        if isinstance(_field_schema, AAZObjectType):
                            # use an empty dict as data for required object property, and process it's sub properties
                            _field_result = processor(_field_schema, {})
                            assert _field_result != AAZUndefined
                            result[_name] = _field_result
                        elif isinstance(_field_schema, AAZBaseDictType):
                            # use an empty dict for required dict property
                            result[_name] = {}
                        elif isinstance(_field_schema, AAZListType):
                            # use an empty dict for required list property
                            result[_name] = []
                        else:
                            raise ValueError(f"Missing a required field in request content: {_name}")

            return result

        if isinstance(value, AAZBaseValue):
            data = value.to_serialized_data(processor=processor)
            flags = value._schema._flags
            required = required or flags.get('required', False) and not flags.get('read_only', False)
            if data == AAZUndefined and required:
                if isinstance(value._schema, AAZObjectType):
                    # use an empty dict as data for required object, and process it's properties
                    data = processor(value._schema, {})
                    assert data != AAZUndefined
                elif isinstance(value._schema, AAZBaseDictType):
                    # use an empty dict for required dict
                    data = {}
                elif isinstance(value._schema, AAZListType):
                    # use an empty list for required list
                    data = []
                else:
                    raise ValueError("Missing request content")
        else:
            data = value

        if data == AAZUndefined or data == None:  # noqa: E711, pylint: disable=singleton-comparison
            if required:
                raise ValueError("Missing request content")
            return None
        return data

    @staticmethod
    def deserialize_http_content(session):
        from azure.core.pipeline.policies import ContentDecodePolicy
        if ContentDecodePolicy.CONTEXT_NAME in session.context:
            return session.context[ContentDecodePolicy.CONTEXT_NAME]
        if session.context.options['stream']:
            # Cannot handle stream response now
            raise NotImplementedError()
        raise ValueError("This pipeline didn't have the ContentDecode Policy; can't deserialize")

    @staticmethod
    def new_content_builder(arg_value, value=None, typ=None, typ_kwargs=None):
        """ Create a Content Builder
        """
        assert isinstance(arg_value, AAZBaseValue)
        arg_data = arg_value.to_serialized_data(keep_undefined_in_list=True)
        if value is None:
            assert issubclass(typ, AAZBaseType)
            schema = typ(**typ_kwargs) if typ_kwargs else typ()

            if isinstance(schema, AAZSimpleType):
                value = typ._ValueCls(
                    schema=schema,
                    data=schema.process_data(arg_data)
                )
            else:
                value = typ._ValueCls(
                    schema=schema,
                    data=schema.process_data(None)
                )
        else:
            assert isinstance(value, AAZBaseValue), f"Unknown value type: {type(value)}"

        builder = AAZContentBuilder(
            values=[value],
            args=[AAZArgBrowser(arg_value=arg_value, arg_data=arg_data)]
        )
        return value, builder

    # properties

    @property
    def url(self):
        return None

    @property
    def method(self):
        return None

    @property
    def url_parameters(self):
        return {}

    @property
    def query_parameters(self):
        return {}

    @property
    def header_parameters(self):
        return {}

    @property
    def content(self):
        return None

    @property
    def form_content(self):
        return None

    @property
    def stream_content(self):
        return None

    @property
    def error_format(self):
        # value should be in the keys of aaz._error_format.registered_error_formats
        return None

    def make_request(self):
        """ Make http request based on the properties.
        """
        if self.ctx.next_link:
            # support making request for next link
            _parsed_next_link = urlparse(self.ctx.next_link)
            _next_request_params = {
                key: [quote(v) for v in value]
                for key, value in parse_qs(_parsed_next_link.query).items()
            }
            request = self.client._request(
                "GET", urljoin(self.ctx.next_link, _parsed_next_link.path), _next_request_params,
                self.header_parameters, self.content, self.form_content, None)

        elif self.method in ("GET",):
            request = self.client._request(
                self.method, self.url, self.query_parameters, self.header_parameters,
                self.content, self.form_content, None)

        elif self.method in ("DELETE", "MERGE", "OPTIONS"):
            request = self.client._request(
                self.method, self.url, self.query_parameters, self.header_parameters,
                self.content, self.form_content, None)

        elif self.method in ("PUT", "POST", "HEAD", "PATCH",):
            request = self.client._request(
                self.method, self.url, self.query_parameters, self.header_parameters,
                self.content, self.form_content, self.stream_content)
        else:
            raise ValueError(f"Invalid request method {self.method}")
        return request

    def on_error(self, response):
        """ handle errors in response
        """
        # raise common http errors
        error_type = self.error_map.get(response.status_code)
        if error_type:
            raise error_type(response=response)
        # raise HttpResponseError
        error_format = self.ctx.get_error_format(self.error_format)
        raise HttpResponseError(response=response, error_format=error_format)


class AAZJsonInstanceOperationHelper:

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()

    @staticmethod
    def new_content_builder(arg_value, value=None, typ=None, typ_kwargs=None):
        """ Create a Content Builder
        """
        assert isinstance(arg_value, AAZBaseValue)
        arg_data = arg_value.to_serialized_data(keep_undefined_in_list=True)
        if value is None:
            assert issubclass(typ, AAZBaseType)
            schema = typ(**typ_kwargs) if typ_kwargs else typ()

            if isinstance(schema, AAZSimpleType):
                value = typ._ValueCls(
                    schema=schema,
                    data=schema.process_data(arg_data)
                )
            else:
                value = typ._ValueCls(
                    schema=schema,
                    data=schema.process_data(None)
                )
        else:
            assert isinstance(value, AAZBaseValue), f"Unknown value type: {type(value)}"

        builder = AAZContentBuilder(
            values=[value],
            args=[AAZArgBrowser(arg_value=arg_value, arg_data=arg_data)]
        )
        return value, builder


class AAZJsonInstanceUpdateOperation(AAZJsonInstanceOperationHelper, AAZOperation):
    """ Instance Update Operation
    """

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()


class AAZJsonInstanceCreateOperation(AAZJsonInstanceOperationHelper, AAZOperation):
    """ Json Instance Create Operation
    """

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()


class AAZJsonInstanceDeleteOperation(AAZJsonInstanceOperationHelper, AAZOperation):
    """ Json Instance Delete Operation
    """

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()

    def _delete_instance(self, *args, **kwargs):  # pylint: disable=unused-argument, no-self-use
        return AAZUndefined


class AAZGenericInstanceUpdateOperation(AAZOperation):
    """ Generic Instance Update Operation
    """

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()

    def _update_instance_by_generic(self, instance, generic_update_args, client_flatten=True):  # pylint: disable=unused-argument
        from azure.cli.core.commands.arm import add_usage, remove_usage, set_usage
        if not generic_update_args or not generic_update_args['actions']:
            return instance
        assert isinstance(instance, AAZBaseValue)

        force_string = generic_update_args.get('force_string', False)
        for action_name, values in generic_update_args['actions']:
            if action_name == "set":
                try:
                    for expression in values:
                        self._set_properties(instance, expression, force_string, client_flatten)
                except ValueError:
                    raise InvalidArgumentValueError('invalid syntax: {}'.format(set_usage))
            elif action_name == "add":
                try:
                    self._add_properties(instance, values, force_string, client_flatten)
                except ValueError:
                    raise InvalidArgumentValueError('invalid syntax: {}'.format(add_usage))
            elif action_name == "remove":
                try:
                    self._remove_properties(instance, values, client_flatten)
                except ValueError:
                    raise InvalidArgumentValueError('invalid syntax: {}'.format(remove_usage))
        return instance

    def _set_properties(self, instance, expression, force_string, flatten):
        from azure.cli.core.commands.arm import _split_key_value_pair, shell_safe_json_parse, index_or_filter_regex
        key, value = _split_key_value_pair(expression)

        if key is None or key.strip() == '':
            raise InvalidArgumentValueError('Empty key in --set. Correct syntax: --set KEY=VALUE [KEY=VALUE ...]')

        if not force_string:
            try:
                value = shell_safe_json_parse(value)
            except (ValueError, InvalidArgumentValueError):
                pass

        path = self._get_internal_path(key)
        name = path.pop()
        instance = self._find_property(instance, path, flatten)  # find the parent instance

        match = index_or_filter_regex.match(name)
        index_value = int(match.group(1)) if match else None
        try:
            if index_value is not None:
                # instance is AAZList or list
                if not isinstance(instance, (AAZList, list)):
                    raise TypeError()
                if len(instance) <= index_value:
                    raise IndexError()
                instance[index_value] = value
            else:
                # instance is in AAZObject or AAZDict or dict
                if isinstance(instance, AAZObject):
                    parent = self._get_property_parent(instance, name, flatten)
                    if parent is None:
                        raise AttributeError()
                    instance = parent
                instance[name] = value
        except AAZInvalidValueError as err:
            raise InvalidArgumentValueError(err)
        except IndexError:
            raise InvalidArgumentValueError('index {} doesn\'t exist on {}'.format(index_value, path[-1]))
        except (AttributeError, KeyError, TypeError):
            self._throw_and_show_options(instance, name, key.split('.'), flatten)

    def _add_properties(self, instance, argument_values, force_string, flatten):
        from azure.cli.core.commands.arm import shell_safe_json_parse
        # The first argument indicates the path to the collection to add to.
        argument_values = list(argument_values)
        list_attribute_path = self._get_internal_path(argument_values.pop(0))
        list_to_add_to = self._find_property(instance, list_attribute_path, flatten)

        if not isinstance(list_to_add_to, (AAZList, list)):
            raise ValueError()

        try:
            dict_entry = {}
            for argument in argument_values:
                if '=' in argument:
                    # consecutive key=value entries get added to the same dictionary
                    split_arg = argument.split('=', 1)
                    argument = split_arg[1]
                    # Didn't convert argument by shell_safe_json_parse here to keep consist with the behavior in arm.py
                    dict_entry[split_arg[0]] = argument
                else:
                    if dict_entry:
                        # if an argument is supplied that is not key=value, append any dictionary entry
                        # to the list and reset. A subsequent key=value pair will be added to another
                        # dictionary.
                        list_to_add_to.append(dict_entry)
                        dict_entry = {}

                    if not force_string:
                        # attempt to convert anything else to JSON and fallback to string if error
                        try:
                            argument = shell_safe_json_parse(argument)
                        except (ValueError, InvalidArgumentValueError):
                            pass
                    list_to_add_to.append(argument)

            # if only key=value pairs used, must check at the end to append the dictionary
            if dict_entry:
                list_to_add_to.append(dict_entry)
        except AAZInvalidValueError as err:
            raise InvalidArgumentValueError(err)

    def _remove_properties(self, instance, argument_values, flatten):
        # The first argument indicates the path to the collection to remove from.
        argument_values = list(argument_values) if isinstance(argument_values, list) else [argument_values]

        list_attribute_path = self._get_internal_path(argument_values.pop(0))
        list_index = None
        try:
            list_index = argument_values.pop(0)
        except IndexError:
            pass

        if not list_index:
            # parent is in AAZObject or AAZDict or dict
            property_val = self._find_property(instance, list_attribute_path, flatten)
            parent = self._find_property(instance, list_attribute_path[:-1], flatten)
            if isinstance(parent, (AAZObject, AAZBaseDictValue)):
                if isinstance(parent, AAZObject):
                    parent = self._get_property_parent(parent, list_attribute_path[-1], flatten)
                if isinstance(property_val, (AAZList, list)):
                    if has_value(property_val):
                        # keep consist with the behavior in arm.py
                        parent[list_attribute_path[-1]] = []
                else:
                    parent[list_attribute_path[-1]] = AAZUndefined
            elif isinstance(parent, dict):
                del parent[list_attribute_path[-1]]
            else:
                # other types
                raise ValueError()
        else:
            list_to_remove_from = self._find_property(instance, list_attribute_path, flatten)
            try:
                if isinstance(list_to_remove_from, AAZList):
                    del list_to_remove_from[int(list_index)]
                elif isinstance(list_to_remove_from, list):
                    list_to_remove_from.pop(int(list_index))
                else:
                    raise IndexError()
            except IndexError:
                raise InvalidArgumentValueError('index {} doesn\'t exist on {}'
                                                .format(list_index, list_attribute_path[-1]))

    def _find_property(self, instance, path, flatten):
        from azure.cli.core.commands.arm import index_or_filter_regex, shell_safe_json_parse
        try:
            for part in path:
                index = index_or_filter_regex.match(part)
                if index and not isinstance(instance, (AAZList, list)):
                    raise AttributeError()

                if index and '=' in index.group(1):
                    key, value = index.group(1).split('=', 1)
                    try:
                        value = shell_safe_json_parse(value)
                    except (ValueError, InvalidArgumentValueError):
                        pass
                    matches = []
                    for x in instance:
                        if isinstance(x, dict) and x.get(key, None) == value:
                            matches.append(x)
                        elif isinstance(x, (AAZObject, AAZBaseDictValue)):
                            parent = x
                            if isinstance(parent, AAZObject):
                                parent = self._get_property_parent(parent, key, flatten)
                                if parent is None:
                                    continue
                            if parent[key].to_serialized_data() == value:
                                matches.append(x)  # should append `x` instead of `parent`

                    if len(matches) > 1:
                        raise InvalidArgumentValueError(
                            "non-unique key '{}' found multiple matches on {}. Key must be unique."
                            .format(key, path[-2]))
                    if len(matches) == 0:
                        raise InvalidArgumentValueError(
                            "item with value '{}' doesn\'t exist for key '{}' on {}".format(value, key, path[-2]))

                    instance = matches[0]

                elif index:
                    try:
                        index_value = int(index.group(1))
                        if isinstance(instance, AAZList):
                            if len(instance) <= index_value:
                                raise IndexError()

                        instance = instance[index_value]
                    except IndexError:
                        raise InvalidArgumentValueError('index {} doesn\'t exist on {}'.format(index_value, path[-2]))

                elif isinstance(instance, (dict, AAZBaseDictValue, AAZObject)):
                    parent = instance
                    if isinstance(instance, AAZObject):
                        parent = self._get_property_parent(parent, part, flatten)
                        if parent is None:
                            raise AttributeError()
                    instance = parent[part]
                else:
                    raise AttributeError()

        except (AttributeError, KeyError):
            self._throw_and_show_options(instance, part, path, flatten)

        return instance

    @staticmethod
    def _get_internal_path(path):
        from azure.cli.core.commands.arm import _get_internal_path
        return _get_internal_path(path)

    def _get_property_parent(self, instance, prop_name, flatten):
        if not isinstance(instance, AAZObject):
            return None

        # prop_name in current instance
        if hasattr(instance, prop_name):
            sub_instance = instance[prop_name]
            if not flatten or not sub_instance._schema._flags.get('client_flatten', False):
                # flatten property should be ignored
                return instance

        if not flatten:
            return None

        for key in self._iter_aaz_object_keys(instance):
            sub_instance = instance[key]
            if sub_instance._schema._flags.get('client_flatten', False):
                sub_instance = self._get_property_parent(sub_instance, prop_name, flatten)
                if sub_instance is not None:
                    return sub_instance

        return None

    @staticmethod
    def _iter_aaz_object_keys(instance):
        if not isinstance(instance, AAZObject):
            return
        schemas = [instance._schema]
        disc_schema = instance._schema.get_discriminator(instance)
        if disc_schema is not None:
            schemas.append(disc_schema)
        for schema in schemas:
            yield from schema._fields

    def _throw_and_show_options(self, instance, part, path, flatten):
        parent = '.'.join(path[:-1]).replace('.[', '[')
        error_message = "Couldn't find '{}' in '{}'.".format(part, parent)
        if isinstance(instance, AAZObject):
            options = []
            for key in self._iter_aaz_object_keys(instance):
                sub_instance = instance[key]
                if flatten and sub_instance._schema._flags.get('client_flatten', False):
                    options.extend(self._iter_aaz_object_keys(sub_instance))
                else:
                    options.append(key)
            options = sorted(options)
            error_message = '{} Available options: {}'.format(error_message, options)
        elif isinstance(instance, (AAZBaseDictValue, dict)):
            options = sorted(instance.keys())
            error_message = '{} Available options: {}'.format(error_message, options)
        elif isinstance(instance, (list, AAZList)):
            options = "index into the collection '{}' with [<index>] or [<key=value>]".format(parent)
            error_message = '{} Available options: {}'.format(error_message, options)
        else:
            error_message = "{} '{}' does not support further indexing.".format(error_message, parent)
        raise InvalidArgumentValueError(error_message)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import base64
from collections import Counter
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from azure.cli.core._profile import Profile
from azure.cli.core.commands.client_factory import get_subscription_id
from azure.cli.command_modules.botservice.web_app_operations import WebAppOperations
from azure.cli.command_modules.botservice.kudu_client import KuduClient


class BotJsonFormatter:  # pylint:disable=too-few-public-methods

    @staticmethod
    def create_bot_json(cmd, client, resource_group_name, resource_name, logger, app_password=None,  # pylint:disable=too-many-locals
                        raw_bot_properties=None, password_only=True):
        """

        :param cmd:
        :param client:
        :param resource_group_name:
        :param resource_name:
        :param logger:
        :param app_password:
        :param raw_bot_properties:
        :return: Dictionary
        """
        if not raw_bot_properties:
            raw_bot_properties = client.bots.get(
                resource_group_name=resource_group_name,
                resource_name=resource_name
            )

        # Initialize names bot_file and secret to capture botFilePath and botFileSecret values from the application's
        # settings.
        bot_file = None
        bot_file_secret = None
        profile = Profile(cli_ctx=cmd.cli_ctx)
        if not app_password:
            site_name = WebAppOperations.get_bot_site_name(raw_bot_properties.properties.endpoint)
            app_settings = WebAppOperations.get_app_settings(
                cmd=cmd,
                resource_group_name=resource_group_name,
                name=site_name
            )

            app_password_values = [item['value'] for item in app_settings if item['name'] == 'MicrosoftAppPassword']
            app_password = app_password_values[0] if app_password_values else None
            if not app_password:
                bot_file_values = [item['value'] for item in app_settings if item['name'] == 'botFilePath']
                bot_file = bot_file_values[0] if bot_file_values else None
                bot_file_secret_values = [item['value'] for item in app_settings if item['name'] == 'botFileSecret']
                bot_file_secret = bot_file_secret_values[0] if bot_file_secret_values else None

        if not bot_file and not app_password:
            bot_site_name = WebAppOperations.get_bot_site_name(raw_bot_properties.properties.endpoint)
            scm_url = WebAppOperations.get_scm_url(cmd,
                                                   resource_group_name,
                                                   bot_site_name,
                                                   None)

            # TODO: Reevaluate "Public-or-Gov" Azure logic.
            is_public_azure = ('azurewebsites.net' in raw_bot_properties.properties.endpoint or
                               '.net' in raw_bot_properties.properties.endpoint or
                               '.com' in raw_bot_properties.properties.endpoint)
            host = 'https://portal.azure.com/' if is_public_azure else 'https://portal.azure.us/'
            subscription_id = get_subscription_id(cmd.cli_ctx)
            tenant_id = profile.get_subscription(subscription=client.config.subscription_id)['tenantId']
            settings_url = host + '#@{}/resource/subscriptions/{}/resourceGroups/{}/providers/Microsoft.BotService/botServices/{}/app_settings'.format(tenant_id, subscription_id, resource_group_name, resource_name)  # pylint: disable=line-too-long

            logger.warning('"MicrosoftAppPassword" and "botFilePath" not found in application settings')
            logger.warning('To see your bot\'s application settings, visit %s' % settings_url)
            logger.warning('To visit your deployed bot\'s code on Azure, visit Kudu for your bot at %s' % scm_url)

        elif not app_password and bot_file:
            # We have the information we need to obtain the MSA App app password via bot file data from Kudu.
            kudu_client = KuduClient(cmd, resource_group_name, resource_name, raw_bot_properties, logger)
            bot_file_data = kudu_client.get_bot_file(bot_file)
            app_password = BotJsonFormatter.__decrypt_bot_file(bot_file_data, bot_file_secret, logger, password_only)

        return {
            'type': 'abs',
            'id': raw_bot_properties.name,
            'name': raw_bot_properties.properties.display_name,
            'appId': raw_bot_properties.properties.msa_app_id,
            'appPassword': app_password,
            'endpoint': raw_bot_properties.properties.endpoint,
            'resourceGroup': str(resource_group_name),
            'tenantId': profile.get_subscription(subscription=client._config.subscription_id)['tenantId'],  # pylint:disable=protected-access
            'subscriptionId': client._config.subscription_id,  # pylint:disable=protected-access
            'serviceName': resource_name
        }

    @staticmethod
    def __decrypt_bot_file(bot_file_data, bot_file_secret, logger, password_only=True):
        """Decrypt .bot file retrieved from Kudu.

        :param bot_file_data:
        :param bot_file_secret:
        :param logger:
        :return:
        """
        services = bot_file_data['services']

        decrypt = BotJsonFormatter.__decrypt

        if password_only:
            # Get all endpoints that have potentially valid appPassword values
            endpoints = [service for service in services
                         if service.get('type') == 'endpoint' and service.get('appPassword')]
            # Reduce the retrieved endpoints to just their passwords
            app_passwords = [e['appPassword'] for e in endpoints]

            if len(app_passwords) == 1:
                return decrypt(bot_file_secret, app_passwords[0], logger)
            if len(app_passwords) > 1:
                logger.info('More than one Microsoft App Password found in bot file. Evaluating if more than one '
                            'unique App Password exists.')
                app_passwords = [decrypt(bot_file_secret, pw, logger) for pw in app_passwords]
                unique_passwords = list(Counter(app_passwords))  # pylint:disable=too-many-function-args
                if len(unique_passwords) == 1:
                    logger.info('One unique Microsoft App Password found, returning password.')
                    return unique_passwords[0]

                logger.warning('More than one unique Microsoft App Password found in the bot file, please '
                               'manually retrieve your bot file from Kudu to retrieve this information.')
                logger.warning('No Microsoft App Password returned.')
                return ''

            logger.warning('No Microsoft App Passwords found in bot file.')
            return ''

        for service in services:
            # For Azure Blob Storage
            if service.get('connectionString'):
                service['connectionString'] = decrypt(bot_file_secret, service['connectionString'], logger)
            # For LUIS and Dispatch
            if service.get('authoringKey'):
                service['authoringKey'] = decrypt(bot_file_secret, service['authoringKey'], logger)
            # For LUIS and QnA Maker
            if service.get('subscriptionKey'):
                service['subscriptionKey'] = decrypt(bot_file_secret, service['subscriptionKey'], logger)
            # For QnA Maker
            if service.get('endpointKey'):
                service['endpointKey'] = decrypt(bot_file_secret, service['endpointKey'], logger)
            # For connecting to the bot
            if service.get('appPassword'):
                service['appPassword'] = decrypt(bot_file_secret, service['appPassword'], logger)
            # For Application Insights
            if service.get('instrumentationKey'):
                service['instrumentationKey'] = decrypt(bot_file_secret, service['instrumentationKey'], logger)
            if service.get('apiKeys'):
                for apiKey in service['apiKeys']:
                    service['apiKeys'][apiKey] = decrypt(bot_file_secret, service['apiKeys'][apiKey], logger)
            # For Cosmos DB
            if service.get('key'):
                service['key'] = decrypt(bot_file_secret, service['key'], logger)
            # For generic services
            if service.get('configuration') and isinstance(service.get('configuration'), dict):
                for key in service['configuration']:
                    service['configuration'][key] = decrypt(bot_file_secret, service['configuration'][key], logger)

        return services

    @staticmethod
    def __decrypt(secret, encrypted_value, logger):
        # If the string length is 0 or no secret was passed in, return the empty string.
        if not encrypted_value or not secret:
            return encrypted_value

        parts = encrypted_value.split("!")
        if len(parts) != 2:
            logger.warn('Encrypted value "%s" not in standard encrypted format, decryption skipped.' % encrypted_value)
            return encrypted_value

        iv_text = parts[0]
        encrypted_text = parts[1]
        iv_bytes = base64.standard_b64decode(str.encode(iv_text))
        secret_bytes = base64.standard_b64decode(str.encode(secret))

        if len(iv_bytes) != 16:
            logger.warn('Initialization Vector for "%s" not valid, decryption skipped.' % encrypted_value)
            return encrypted_value
        if len(secret_bytes) != 32:
            logger.warn('Passed in secret length is invalid, decryption skipped.')
            return encrypted_value

        cipher = Cipher(algorithms.AES(secret_bytes), modes.CBC(iv_bytes), backend=default_backend())
        decryptor = cipher.decryptor()
        decrypted_bytes = decryptor.update(base64.standard_b64decode(str.encode(encrypted_text))) + decryptor.finalize()

        decrypted_string = decrypted_bytes.decode('utf-8')
        return ''.join([char for char in decrypted_string if ord(char) > 31])



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# AZURE CLI EventHub - NAMESPACE TEST DEFINITIONS

import time

from azure.cli.testsdk import (ScenarioTest, ResourceGroupPreparer)


# pylint: disable=line-too-long
# pylint: disable=too-many-lines


class EHNetworkCURDScenarioTest(ScenarioTest):

    @ResourceGroupPreparer(name_prefix='cli_test_eh_network')
    def test_eh_privateendpoint(self, resource_group):
        self.kwargs.update({
            'namespacename': self.create_random_name(prefix='eventhubs-nscli', length=20),
            'loc': 'eastus',
            'tags': {'tag1=value1'},
            'tags2': {'tag2=value2'},
            'sku': 'Standard',
            'vnet': self.create_random_name('cli-vnet-', 24),
            'subnet': self.create_random_name('cli-subnet-', 24),
            'pe': self.create_random_name('cli-pe-', 24),
            'pe_connection': self.create_random_name('cli-pec-', 24),
        })

        # Prepare network
        self.cmd('network vnet create -n {vnet} -g {rg} -l {loc} --subnet-name {subnet}',
                 checks=self.check('length(newVNet.subnets)', 1))
        self.cmd('network vnet subnet update -n {subnet} --vnet-name {vnet} -g {rg} '
                 '--disable-private-endpoint-network-policies true',
                 checks=self.check('privateEndpointNetworkPolicies', 'Disabled'))

        # Create Namespace
        self.cmd(
            'eventhubs namespace create --resource-group {rg} --name {namespacename} --location {loc} --tags {tags}'
            ' --sku {sku}', checks=[self.check('sku.name', '{sku}')])

        # Get Created Namespace
        getnamespace = self.cmd(
            'eventhubs namespace show --resource-group {rg} --name {namespacename}').get_output_in_json()

        # Create a private endpoint connection
        pr = self.cmd('eventhubs namespace private-link-resource show --namespace-name {namespacename} -g {rg}').get_output_in_json()
        self.kwargs['group_id'] = pr[0]['groupId']

        getnamesapce = self.cmd('eventhubs namespace show -n {namespacename} -g {rg}').get_output_in_json()
        self.kwargs['ehn_id'] = getnamesapce['id']
        private_endpoint = self.cmd(
            'network private-endpoint create -g {rg} -n {pe} --vnet-name {vnet} --subnet {subnet} -l {loc} '
            '--connection-name {pe_connection} --private-connection-resource-id {ehn_id} '
            '--group-ids {group_id}').get_output_in_json()
        self.assertEqual(private_endpoint['name'], self.kwargs['pe'])
        self.assertEqual(private_endpoint['privateLinkServiceConnections'][0]['name'], self.kwargs['pe_connection'])
        self.assertEqual(
            private_endpoint['privateLinkServiceConnections'][0]['privateLinkServiceConnectionState']['status'],
            'Approved')
        self.assertEqual(private_endpoint['privateLinkServiceConnections'][0]['provisioningState'], 'Succeeded')
        self.assertEqual(private_endpoint['privateLinkServiceConnections'][0]['groupIds'][0], self.kwargs['group_id'])
        self.kwargs['pe_id'] = private_endpoint['privateLinkServiceConnections'][0]['id']

        # Show the connection at eventhubs namespace
        getnamesapce = self.cmd('eventhubs namespace show -n {namespacename} -g {rg}').get_output_in_json()
        self.assertIn('privateEndpointConnections', getnamesapce)
        self.assertEqual(len(getnamesapce['privateEndpointConnections']), 1)
        self.assertEqual(getnamesapce['privateEndpointConnections'][0]['privateLinkServiceConnectionState']['status'],
                         'Approved')

        self.kwargs['sa_pec_id'] = getnamesapce['privateEndpointConnections'][0]['id']
        self.kwargs['ehn_pec_name'] = getnamesapce['privateEndpointConnections'][0]['name']

        self.cmd('eventhubs namespace private-endpoint-connection show --namespace-name {namespacename} -g {rg} --name {ehn_pec_name}',
                 checks=self.check('id', '{sa_pec_id}'))

        getstatus = self.cmd('eventhubs namespace private-endpoint-connection approve --namespace-name {namespacename} -g {rg} --name {ehn_pec_name}').get_output_in_json()
        self.assertEqual(getstatus['privateLinkServiceConnectionState']['status'], 'Approved')

        getstatus = self.cmd(
            'eventhubs namespace private-endpoint-connection reject --namespace-name {namespacename} -g {rg} --name {ehn_pec_name}').get_output_in_json()
        self.assertEqual(getstatus['privateLinkServiceConnectionState']['status'], 'Rejected')

        getstatus = self.cmd(
            'eventhubs namespace private-endpoint-connection show --namespace-name {namespacename} -g {rg} --name {ehn_pec_name}').get_output_in_json()
        self.assertEqual(getstatus['privateLinkServiceConnectionState']['status'], 'Rejected')

        while getstatus['provisioningState'] != 'Succeeded':
            time.sleep(30)
            getstatus = self.cmd(
                'eventhubs namespace private-endpoint-connection show --namespace-name {namespacename} -g {rg} --name {ehn_pec_name}').get_output_in_json()

        self.cmd('eventhubs namespace private-endpoint-connection delete --id {sa_pec_id} -y')


# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest


class TestAAZField(unittest.TestCase):

    def test_aaz_model_and_simple_types(self):
        from azure.cli.core.aaz._field_type import AAZObjectType, AAZIntType, AAZStrType, AAZBoolType, AAZFloatType
        from azure.cli.core.aaz._field_value import AAZObject
        from azure.cli.core.aaz.exceptions import AAZInvalidValueError
        model_schema = AAZObjectType()
        v = AAZObject(model_schema, data={})

        model_schema.properties = AAZObjectType(options=["prop", "pr"])

        # test int type
        model_schema.properties.count = AAZIntType()
        v.properties.count = 10
        assert isinstance(v.properties, AAZObject)
        assert v.properties._is_patch
        assert not v.properties.count._is_patch
        assert v.properties.count._data == 10
        assert v.properties.count == 10 and v.properties.count <= 10 and v.properties.count >= 10
        assert not (v.properties.count != 10) and not (v.properties.count < 10) and not (v.properties.count > 10)
        assert v.properties.count != 11 and v.properties.count < 11 and v.properties.count <= 11
        assert not (v.properties.count == 11) and not (v.properties.count > 11) and not (v.properties.count >= 11)
        assert v.properties.count != 9 and v.properties.count > 9 and v.properties.count >= 9
        assert not (v.properties.count == 9) and not (v.properties.count < 9) and not (v.properties.count <= 9)
        assert v.properties.count
        assert (not v.properties.count) is False
        v.properties.count = 0
        assert (not v.properties.count) is True

        with self.assertRaises(AAZInvalidValueError):
            v.properties.count = "a"
        
        with self.assertRaises(AAZInvalidValueError):
            v.properties.count = "1a"
        
        with self.assertRaises(AAZInvalidValueError):
            v.properties.count = "1.1"

        # support string input with json parsing
        v.properties.count = "12"
        assert v.properties.count == 12
        assert v.properties.count._data == 12

        # test string type
        model_schema.properties.name = AAZStrType()
        v.properties.name = "a"
        assert not v.properties.name._is_patch
        assert v.properties.name._data == "a"
        assert v.properties.name == "a"
        assert not v.properties.name != "a"
        assert v.properties.name != "b"
        assert not v.properties.name == "b"
        assert v.properties.name
        assert (not v.properties.name) is False
        v.properties.name = ""
        assert (not v.properties.name) is True

        with self.assertRaises(AAZInvalidValueError):
            v.properties.name = 1

        # test bool type
        model_schema.properties.enable = AAZBoolType()
        v.properties.enable = True
        assert not v.properties.enable._is_patch
        assert v.properties.enable == True
        assert not (v.properties.enable != True)
        assert v.properties.enable
        assert v.properties.enable is not True  # cannot us is

        v.properties.enable = False
        assert v.properties.enable == False
        assert not (v.properties.enable != False)
        assert not v.properties.enable
        assert v.properties.enable is not False

        with self.assertRaises(AAZInvalidValueError):
            v.properties.enable = 1

        # support string input with json parsing
        v.properties.enable = "false"
        v.properties.enable = "true"
        assert v.properties.enable == True
        assert not (v.properties.enable != True)
        assert v.properties.enable
        assert v.properties.enable is not True  # cannot us is

        # test float type
        model_schema.properties.height = AAZFloatType()
        v.properties.height = 10.0
        assert str(v.properties.height) == "10.0"
        assert not v.properties.height._is_patch
        assert v.properties.height._data == 10
        assert v.properties.height == 10 and v.properties.height <= 10 and v.properties.height >= 10
        assert not (v.properties.height != 10) and not (v.properties.height < 10) and not (v.properties.height > 10)
        assert v.properties.height != 11 and v.properties.height < 11 and v.properties.height <= 11
        assert not (v.properties.height == 11) and not (v.properties.height > 11) and not (v.properties.height >= 11)
        assert v.properties.height != 9 and v.properties.height > 9 and v.properties.height >= 9
        assert not (v.properties.height == 9) and not (v.properties.height < 9) and not (v.properties.height <= 9)

        v.properties.height = 10  # test assign int
        assert str(v.properties.height) == "10.0"

        with self.assertRaises(AAZInvalidValueError):
            v.properties.height = "1a"
        
        v.properties.height = "12.1"
        assert v.properties.height == 12.1
        assert v.properties.height._data == 12.1

        v.properties.height = "12"
        assert v.properties.height == 12
        assert v.properties.height._data == 12

        # test assign properties by dict
        v.properties = {
            "count": 100,
            "name": "abc",
            "enable": True,
            "height": 111
        }

        assert not v.properties._is_patch
        assert v.properties.name._data == "abc"
        assert v.properties.count._data == 100
        assert v.properties.enable._data == True
        assert v.properties.height._data == 111

        # test assign properties by dict with all string value
        v.properties = {
            "count": "101",
            "name": "abcd",
            "enable": "False",
            "height": "121"
        }

        assert not v.properties._is_patch
        assert v.properties.name._data == "abcd"
        assert v.properties.count._data == 101
        assert v.properties.enable._data == False
        assert v.properties.height._data == 121


    def test_aaz_any_type(self):
        from azure.cli.core.aaz._field_type import AAZObjectType, AAZDictType, AAZAnyType, AAZListType, AAZIntType, AAZStrType, AAZBoolType
        from azure.cli.core.aaz._field_value import AAZObject, AAZAnyValue, AAZValuePatch, AAZUndefined
        model_schema = AAZObjectType()
        model_schema.any = AAZAnyType()
        model_schema.nullable_any = AAZAnyType(nullable=True)
        model_schema.additional = AAZDictType()
        model_schema.additional.Element = AAZAnyType()
        model_schema.nullable_additional = AAZDictType()
        model_schema.nullable_additional.Element = AAZAnyType(nullable=True)
        model_schema.array = AAZListType()
        model_schema.array.Element = AAZAnyType()
        model_schema.nullable_array = AAZListType()
        model_schema.nullable_array.Element = AAZAnyType(nullable=True)
        model_schema.patch_object = AAZObjectType(nullable=True)
        model_schema.patch_object.obj = AAZObjectType()
        model_schema.patch_object.obj.a = AAZIntType()
        model_schema.patch_object.obj.b = AAZStrType()
        model_schema.patch_object.obj.c = AAZBoolType()
        model_schema.patch_object.obj.d = AAZAnyType(nullable=True)

        v = AAZObject(schema=model_schema, data={})
        assert v.any._is_patch
        self.assertEqual(v.to_serialized_data(), {})

        v.any = 1
        self.assertEqual(v.any.to_serialized_data(), 1)
        v.any = 0.1
        self.assertEqual(v.any.to_serialized_data(), 0.1)
        v.any = "1"
        self.assertEqual(v.any.to_serialized_data(), "1")
        v.any = True
        self.assertEqual(v.any.to_serialized_data(), True)
        v.any = False
        self.assertEqual(v.any.to_serialized_data(), False)
        v.any = [1, "123", None]
        self.assertEqual(v.any.to_serialized_data(), [1, "123", None])
        v.any = None
        self.assertEqual(v.to_serialized_data(), {})
        v.nullable_any = None
        self.assertEqual(v.to_serialized_data(), {"nullable_any": None})

        # verify additional and nullable_additional
        v.additional = {"a": 1, "b": "str", "c": True, "d": None}
        self.assertEqual(v.additional.to_serialized_data(), {"a": 1, "b": "str", "c": True})
        v.nullable_additional = {"a": 1, "b": "str", "c": True, "d": None}
        self.assertEqual(v.nullable_additional.to_serialized_data(), {"a": 1, "b": "str", "c": True, "d": None})

        # verify array and nullable_array
        v.array = [1, "123", None]
        self.assertEqual(v.array.to_serialized_data(), [1, "123"])
        v.nullable_array = [1, "123", None]
        self.assertEqual(v.nullable_array.to_serialized_data(), [1, "123", None])

        # verify the patch value assign to any type
        # undefined patch value
        _ = v.patch_object.obj.a
        self.assertEqual(v.patch_object.obj.a._is_patch, True)
        
        v.any = v.patch_object
        self.assertEqual(v.any._is_patch, True)
        self.assertEqual(v.any.to_serialized_data(), AAZUndefined)
        
        v.array[2] = v.patch_object
        self.assertEqual(v.array[2]._is_patch, True)
        self.assertEqual(v.array.to_serialized_data(), [1, "123"])
        
        v.nullable_array[3] = v.patch_object
        self.assertEqual(v.nullable_array[3]._is_patch, True)
        self.assertEqual(v.nullable_array.to_serialized_data(), [1, "123", None])

        # defined patch value
        v.patch_object.obj.a = 1
        v.patch_object.obj.b = "str"

        v.any = v.patch_object
        self.assertEqual(v.any._is_patch, False)
        self.assertEqual(v.any.to_serialized_data(), {"obj": {"a": 1, "b": "str"}})
        
        v.array[2] = v.patch_object
        self.assertEqual(v.array[2]._is_patch, False)
        self.assertEqual(v.array.to_serialized_data(), [1, "123", {"obj": {"a": 1, "b": "str"}}])
        
        v.nullable_array[3] = v.patch_object
        self.assertEqual(v.nullable_array[3]._is_patch, False)
        self.assertEqual(v.nullable_array.to_serialized_data(), [1, "123", None, {"obj": {"a": 1, "b": "str"}}])

        # defined patch value is None
        v.patch_object = None
        
        v.any = v.patch_object
        self.assertEqual(v.any._is_patch, True)
        self.assertEqual(v.any.to_serialized_data(), AAZUndefined)
        
        v.nullable_any = v.patch_object
        self.assertEqual(v.nullable_any._is_patch, False)
        self.assertEqual(v.nullable_any.to_serialized_data(), None)

         
        v.array[2] = v.patch_object
        self.assertEqual(v.array[2]._is_patch, True)
        self.assertEqual(v.array.to_serialized_data(), [1, "123"])
        
        v.nullable_array[3] = v.patch_object
        self.assertEqual(v.nullable_array[3]._is_patch, False)
        self.assertEqual(v.nullable_array.to_serialized_data(), [1, "123", None, None])


    def test_aaz_dict_type(self):
        from azure.cli.core.aaz._field_type import AAZObjectType, AAZDictType, AAZStrType
        from azure.cli.core.aaz._field_value import AAZObject
        model_schema = AAZObjectType()
        model_schema.tags = AAZDictType()
        model_schema.tags.Element = AAZStrType()
        model_schema.additional = AAZDictType()
        model_schema.additional.Element = AAZObjectType()
        model_schema.additional.Element.name = AAZStrType()

        v = AAZObject(schema=model_schema, data={})
        assert v.tags["p"]._is_patch
        del v.tags["p"]
        with self.assertRaises(KeyError):
            del v.tags["a"]

        v.tags["flag_1"] = "f1"

        assert v.tags["flag_1"]._data == "f1"
        assert v.tags["flag_1"] == "f1"
        assert "flag_1" in v.tags
        assert "flag_2" not in v.tags
        assert len(v.tags) == 1

        for key in v.tags:
            assert key == "flag_1"

        for key in v.tags.keys():
            assert key == "flag_1"

        for value in v.tags.values():
            assert value._data == "f1"
            assert value == "f1"

        for key, value in v.tags.items():
            assert key == "flag_1"
            assert value._data == "f1"
            assert value == "f1"

        v.tags.clear()
        assert len(v.tags) == 0
        for key in v.tags:
            assert False
        for key in v.tags.keys():
            assert False
        for value in v.tags.values():
            assert False
        for key, value in v.tags.items():
            assert False

        del v.tags
        for key in v.tags:
            assert False
        del v.tags
        for key in v.tags.keys():
            assert False
        del v.tags
        for value in v.tags.values():
            assert False
        del v.tags
        for key, value in v.tags.items():
            assert False

    def test_aaz_free_form_dict_type(self):
        from azure.cli.core.aaz._field_type import AAZObjectType, AAZFreeFormDictType, AAZDictType, AAZListType
        from azure.cli.core.aaz._field_value import AAZObject, AAZBaseValue, AAZUndefined, AAZValuePatch, AAZAnyValue
        from azure.cli.core.aaz.exceptions import AAZInvalidValueError

        model_schema = AAZObjectType()
        model_schema.additional = AAZFreeFormDictType()
        model_schema.configs = AAZListType()
        model_schema.configs.Element = AAZFreeFormDictType()
        model_schema.nullable_additional = AAZFreeFormDictType(nullable=True)

        v = AAZObject(schema=model_schema, data={})

        assert 'A' not in v.additional
        assert v.additional['A']._is_patch
        del v.additional['A']
        with self.assertRaises(KeyError):
            del v.additional['A']

        v.additional['number'] = 1
        self.assertEqual(v.additional['number'], 1)

        for key in v.additional:
            self.assertEqual(key, 'number')

        for key in v.additional.keys():
            self.assertEqual(key, "number")

        for key, value in v.additional.items():
            self.assertEqual(key, 'number')
            self.assertEqual(value, 1)
            self.assertTrue(isinstance(value, AAZAnyValue))

        v.additional.clear()
        self.assertEqual(len(v.additional), 0)

        for key in v.additional:
            self.assertTrue(False)

        for key in v.additional.keys():
            self.assertTrue(False)

        for key, value in v.additional.items():
            self.assertTrue(False)

        v.additional['str'] = 'str'
        self.assertEqual(v.additional['str'], 'str')

        del v.additional

        for key in v.additional:
            self.assertTrue(False)

        for key in v.additional.keys():
            self.assertTrue(False)

        for key, value in v.additional.items():
            self.assertTrue(False)

        v.additional['bool'] = True
        self.assertEqual(v.additional['bool'], True)

        del v.additional['bool']
        with self.assertRaises(KeyError):
            del v.additional['bool']

        v.additional['none'] = None
        self.assertEqual(v.additional['none'], None)

        v.additional['obj'] = {"a": 1, "b": "str", "c": True, "d": None}
        self.assertEqual(v.additional['obj'], {"a": 1, "b": "str", "c": True, "d": None})

        v.additional['list'] = ['a', 1, True, None, ["s", "v"], {"a": 1, "b": 2}]
        self.assertEqual(v.additional['list'], ['a', 1, True, None, ["s", "v"], {"a": 1, "b": 2}])

        self.assertTrue(v.additional._is_patch)

        self.assertEqual(v.to_serialized_data(), {
            "additional": {
                'none': None,
                'obj': {'a': 1, 'b': 'str', 'c': True, 'd': None},
                'list': ['a', 1, True, None, ['s', 'v'], {'a': 1, 'b': 2}]
            }
        })

        v.configs[0] = {'a': 1, 'b': 'str', 'c': True, 'd': None}
        self.assertEqual(v.configs[0].to_serialized_data(), {'a': 1, 'b': 'str', 'c': True, 'd': None})
        self.assertEqual(v.configs[0]._data, {'a': 1, 'b': 'str', 'c': True, 'd': None})

        v.configs[1] = {"obj": {"a": 1, "c": 2}}
        # AAZAnyValue is not support delete
        with self.assertRaises(TypeError):
            del v.configs[1]['obj']['a']
        self.assertEqual(v.configs[1].to_serialized_data(), {"obj": {"a": 1, "c": 2}})
        self.assertEqual(v.configs[1]._data, {"obj": {"a": 1, "c": 2}})
        v.configs[1] = {"obj": {"c": 2}}

        v.configs[0] = None
        self.assertEqual(v.configs.to_serialized_data(), [{"obj": {"c": 2}}])

        v.nullable_additional['a'] = AAZValuePatch(AAZUndefined)
        self.assertTrue(v.nullable_additional._is_patch)

        self.assertTrue(v.additional._is_patch)
        v.nullable_additional['p'] = v.additional

        self.assertTrue(v.configs[1]._is_patch is False)

        v.nullable_additional['z'] = v.configs[1]

        v.configs[1]['more'] = True
        self.assertEqual(v.nullable_additional.to_serialized_data(), {
            "p": {
                'none': None,
                'obj': {'a': 1, 'b': 'str', 'c': True, 'd': None},
                'list': ['a', 1, True, None, ['s', 'v'], {'a': 1, 'b': 2}]
            },
            "z": {"obj": {"c": 2}},
        })

        self.assertEqual(v.to_serialized_data(), {
            "additional": {
                'none': None,
                'obj': {'a': 1, 'b': 'str', 'c': True, 'd': None},
                'list': ['a', 1, True, None, ['s', 'v'], {'a': 1, 'b': 2}]
            },
            "configs": [{"obj": {"c": 2}, "more": True}],
            "nullable_additional": {
                "p": {
                    'none': None,
                    'obj': {'a': 1, 'b': 'str', 'c': True, 'd': None},
                    'list': ['a', 1, True, None, ['s', 'v'], {'a': 1, 'b': 2}]
                },
                "z": {"obj": {"c": 2}}}
        })

        v.nullable_additional = None
        self.assertEqual(v.to_serialized_data(), {
            "additional": {
                'none': None,
                'obj': {'a': 1, 'b': 'str', 'c': True, 'd': None},
                'list': ['a', 1, True, None, ['s', 'v'], {'a': 1, 'b': 2}]
            },
            "configs": [{"obj": {"c": 2}, "more": True}],
            "nullable_additional": None
        })

    def test_aaz_list_type(self):
        from azure.cli.core.aaz._field_type import AAZObjectType, AAZListType, AAZStrType
        from azure.cli.core.aaz._field_value import AAZObject
        model_schema = AAZObjectType()
        model_schema.sub_nets = AAZListType()
        model_schema.sub_nets.Element = AAZObjectType()
        model_schema.sub_nets.Element.id = AAZStrType()

        v = AAZObject(schema=model_schema, data={})
        v.sub_nets[0].id = "/0/"
        assert len(v.sub_nets) == 1
        assert v.sub_nets[0]._is_patch
        assert v.sub_nets[0].id._data == "/0/"
        assert v.sub_nets[0].id == "/0/"
        v.sub_nets[1] = {
            "id": "/1/"
        }
        assert len(v.sub_nets) == 2
        assert not v.sub_nets[1]._is_patch
        assert v.sub_nets[1].id._data == "/1/"
        assert v.sub_nets[1].id == "/1/"

        assert not v.sub_nets[-1]._is_patch
        assert v.sub_nets[-1].id._data == "/1/"
        assert v.sub_nets[-1].id == "/1/"

        assert v.sub_nets._is_patch

        v.sub_nets = [
            {
                "id": "/2/",
            },
        ]

        assert len(v.sub_nets) == 1
        assert not v.sub_nets._is_patch
        assert not v.sub_nets[0]._is_patch
        assert v.sub_nets[0].id._data == "/2/"
        assert v.sub_nets[0].id == "/2/"

        for sub_net in v.sub_nets:
            assert sub_net.id == "/2/"

        v.sub_nets.append({
            "id": "/3/"
        })
        assert len(v.sub_nets) == 2
        v.sub_nets.clear()
        assert len(v.sub_nets) == 0
        for sub_net in v.sub_nets:
            assert False

        v.sub_nets.extend([{"id": "/1/"}, {"id": "/2/"}])
        assert len(v.sub_nets) == 2

        del v.sub_nets
        for sub_net in v.sub_nets:
            assert False

    def test_aaz_types_with_alias(self):
        from azure.cli.core.aaz._field_type import AAZObjectType, AAZStrType, AAZDictType, AAZListType
        from azure.cli.core.aaz._field_value import AAZObject
        model_schema = AAZObjectType()
        model_schema.sub_properties = AAZObjectType(options=["sub-properties"], serialized_name="subProperties")
        model_schema.sub_properties.vnet_id = AAZStrType(options=["vnet-id"], serialized_name="vnetId")
        model_schema.sub_properties.sub_tags = AAZDictType(options=["sub-tags"], serialized_name="subTags")
        model_schema.sub_properties.sub_tags.Element = AAZStrType()
        model_schema.sub_properties.sub_nets = AAZListType(options=["sub-nets"], serialized_name="subNets")
        model_schema.sub_properties.sub_nets.Element = AAZObjectType()
        model_schema.sub_properties.sub_nets.Element.address_id = AAZStrType(options=["address-id"],
                                                                             serialized_name="addressId")

        v = AAZObject(model_schema, data={})

        v.sub_properties = {
            "vnetId": "vnetId",
            "subTags": {
                "aa": "aa",
                "bb": "bb"
            },
            "subNets": [
                {
                    "addressId": "1.1.1.1"
                },
                {
                    "addressId": "2.2.2.2"
                }
            ]
        }
        assert v.sub_properties.vnet_id._data == "vnetId"
        assert v.sub_properties.sub_tags._data == {"aa": "aa", "bb": "bb"}
        assert v.sub_properties.sub_nets[0].address_id._data == "1.1.1.1"
        assert v.sub_properties.sub_nets[1].address_id._data == "2.2.2.2"
        assert v.sub_properties.vnetId._data == "vnetId"
        assert v.sub_properties.subTags._data == {"aa": "aa", "bb": "bb"}
        assert v.sub_properties.subNets[0].addressId._data == "1.1.1.1"
        assert v.sub_properties.subNets[1].addressId._data == "2.2.2.2"

        v.sub_properties = {
            "vnet-id": "vnet-id",
            "sub-tags": {
                "cc": "cc",
            },
            "sub-nets": [
                {
                    "address-id": "3.3.3.3"
                }
            ]
        }

        assert v.sub_properties.vnet_id._data == "vnet-id"
        assert v.sub_properties.sub_tags._data == {"cc": "cc"}
        assert v.sub_properties.sub_nets[0].address_id._data == "3.3.3.3"
        assert v.sub_properties.vnetId._data == "vnet-id"
        assert v.sub_properties.subTags._data == {"cc": "cc"}
        assert v.sub_properties.subNets[0].addressId._data == "3.3.3.3"

        model_schema.sub_properties.err_sub_nets = AAZListType()
        with self.assertRaises(AssertionError):
            model_schema.sub_properties.err_sub_nets.Element = AAZStrType(options=["aaa"])

        model_schema.sub_properties.err_sub_tags = AAZDictType()
        with self.assertRaises(AssertionError):
            model_schema.sub_properties.err_sub_tags.Element = AAZStrType(options=["bbb"])

    def test_aaz_recursive_type(self):
        from azure.cli.core.aaz._field_type import AAZObjectType, AAZStrType, AAZDictType, AAZListType
        from azure.cli.core.aaz._field_value import AAZObject
        model_schema = AAZObjectType()

        v = AAZObject(model_schema, data={})

        model_schema.errors = AAZListType()

        error = model_schema.errors.Element = AAZObjectType()

        error.code = AAZStrType()
        error.message = AAZStrType()
        error.target = AAZStrType()
        error.additional_info = AAZListType(serialized_name="additionalInfo")
        error.details = AAZListType()
        error.details.Element = error

        add_info_element = model_schema.errors.Element.additional_info.Element = AAZObjectType()
        add_info_element.type = AAZStrType()
        add_info_element.info = AAZDictType()
        add_info_element.info.Element = AAZStrType()

        v.errors = [
            {
                "code": "",
                "message": "",
                "target": "",
                "additionalInfo": [],
                "details": [
                    {
                        "code": "",
                        "target": "",
                        "message": "",
                        "additionalInfo": [
                            {
                                "type": "PolicyViolation",
                                "info": {
                                    "policySetDefinitionDisplayName": "Secure the environment",
                                    "policySetDefinitionId": "/subscriptions/00000-00000-0000-000/providers/Microsoft.Authorization/policySetDefinitions/TestPolicySet",
                                    "policyDefinitionDisplayName": "Allowed locations",
                                    "policyDefinitionId": "/subscriptions/00000-00000-0000-000/providers/Microsoft.Authorization/policyDefinitions/TestPolicy1",
                                    "policyAssignmentDisplayName": "Allow Central US and WEU only",
                                    "policyAsssignmentId": "/subscriptions/00000-00000-0000-000/providers/Microsoft.Authorization/policyAssignments/TestAssignment1"
                                }
                            },
                            {
                                "type": "SomeOtherViolation",
                                "info": {
                                    "innerException": "innerException Details"
                                }
                            }
                        ],
                        "details": [
                            {
                                "code": "",
                                "target": "",
                                "message": "",
                                "additionalInfo": [
                                    {
                                        "type": "PolicyViolation",
                                        "info": {
                                            "policySetDefinitionDisplayName": "Secure the environment",
                                            "policySetDefinitionId": "/subscriptions/00000-00000-0000-000/providers/Microsoft.Authorization/policySetDefinitions/TestPolicySet",
                                            "policyDefinitionDisplayName": "Allowed locations",
                                            "policyDefinitionId": "/subscriptions/00000-00000-0000-000/providers/Microsoft.Authorization/policyDefinitions/TestPolicy1",
                                            "policyAssignmentDisplayName": "Allow Central US and WEU only",
                                            "policyAsssignmentId": "/subscriptions/00000-00000-0000-000/providers/Microsoft.Authorization/policyAssignments/TestAssignment1"
                                        }
                                    },
                                    {
                                        "type": "SomeOtherViolation",
                                        "info": {
                                            "innerException": "innerException Details"
                                        }
                                    }
                                ],
                            }
                        ]
                    }
                ]
            }
        ]

        data = v.to_serialized_data()
        self.assertTrue(data == {
            "errors": [
                {
                    "code": "",
                    "message": "",
                    "target": "",
                    "additionalInfo": [],
                    "details": [
                        {
                            "code": "",
                            "target": "",
                            "message": "",
                            "additionalInfo": [
                                {
                                    "type": "PolicyViolation",
                                    "info": {
                                        "policySetDefinitionDisplayName": "Secure the environment",
                                        "policySetDefinitionId": "/subscriptions/00000-00000-0000-000/providers/Microsoft.Authorization/policySetDefinitions/TestPolicySet",
                                        "policyDefinitionDisplayName": "Allowed locations",
                                        "policyDefinitionId": "/subscriptions/00000-00000-0000-000/providers/Microsoft.Authorization/policyDefinitions/TestPolicy1",
                                        "policyAssignmentDisplayName": "Allow Central US and WEU only",
                                        "policyAsssignmentId": "/subscriptions/00000-00000-0000-000/providers/Microsoft.Authorization/policyAssignments/TestAssignment1"
                                    }
                                },
                                {
                                    "type": "SomeOtherViolation",
                                    "info": {
                                        "innerException": "innerException Details"
                                    }
                                }
                            ],
                            "details": [
                                {
                                    "code": "",
                                    "target": "",
                                    "message": "",
                                    "additionalInfo": [
                                        {
                                            "type": "PolicyViolation",
                                            "info": {
                                                "policySetDefinitionDisplayName": "Secure the environment",
                                                "policySetDefinitionId": "/subscriptions/00000-00000-0000-000/providers/Microsoft.Authorization/policySetDefinitions/TestPolicySet",
                                                "policyDefinitionDisplayName": "Allowed locations",
                                                "policyDefinitionId": "/subscriptions/00000-00000-0000-000/providers/Microsoft.Authorization/policyDefinitions/TestPolicy1",
                                                "policyAssignmentDisplayName": "Allow Central US and WEU only",
                                                "policyAsssignmentId": "/subscriptions/00000-00000-0000-000/providers/Microsoft.Authorization/policyAssignments/TestAssignment1"
                                            }
                                        },
                                        {
                                            "type": "SomeOtherViolation",
                                            "info": {
                                                "innerException": "innerException Details"
                                            }
                                        }
                                    ],
                                }
                            ]
                        }
                    ]
                }
            ]
        })

    def test_aaz_object_with_polymorphism_support(self):
        from azure.cli.core.aaz._field_type import AAZObjectType, AAZListType, AAZStrType, AAZIntType
        from azure.cli.core.aaz._field_value import AAZObject
        from azure.cli.core.aaz.exceptions import AAZUnknownFieldError
        from azure.cli.core.aaz._command import AAZCommand

        model_schema = AAZObjectType()
        v = AAZObject(model_schema, data={})

        model_schema.display_name = AAZStrType()
        model_schema.actions = AAZListType()

        element = model_schema.actions.Element = AAZObjectType()

        element.action_type = AAZStrType()
        element.order = AAZIntType()
        disc_modify_properties = element.discriminate_by('action_type', 'ModifyProperties')
        action_configuration = disc_modify_properties.action_configuration = AAZObjectType(
            flags={"client_flatten": True})
        action_configuration.classification = AAZStrType()
        action_configuration.classification_comment = AAZStrType()
        action_configuration.severity = AAZStrType()
        action_configuration.secret = AAZStrType(flags={"secret": True})

        disc_run_playbook = element.discriminate_by('action_type', 'RunPlaybook')
        disc_run_playbook.logic_app_resource_id = AAZStrType()
        disc_run_playbook.tenant_id = AAZStrType()

        v.actions = [
            {
                "action_type": "ModifyProperties",
                "order": 0,
                "action_configuration": {
                    "classification": "BenignPositive",
                    "classification_comment": "comments 1",
                    "severity": "High",
                    "secret": "secret-value",
                }
            },
            {
                "action_type": "RunPlaybook",
                "order": 1,
                "logic_app_resource_id": "123333",
                "tenant_id": "111111",
            }
        ]

        self.assertTrue(v.to_serialized_data() == {
            "actions": [
                {
                    "action_type": "ModifyProperties",
                    "order": 0,
                    "action_configuration": {
                        "classification": "BenignPositive",
                        "classification_comment": "comments 1",
                        "severity": "High",
                        "secret": "secret-value",
                    }
                },
                {
                    "action_type": "RunPlaybook",
                    "order": 1,
                    "logic_app_resource_id": "123333",
                    "tenant_id": "111111",
                }
            ]
        })

        v.actions[2] = {
            "action_type": "ModifyProperties",
        }

        v.actions[2].order = 2
        with self.assertRaises(AAZUnknownFieldError):
            v.actions[2].logic_app_resource_id = "6666"
        self.assertTrue(v.actions[2].action_configuration.classification._is_patch)
        v.actions[2].action_configuration.classification = "FalsePositive"
        self.assertEqual(v.actions[2].action_configuration.classification, "FalsePositive")
        self.assertEqual(v.actions[2].actionConfiguration.classification, "FalsePositive")

        self.assertTrue(v.to_serialized_data() == {
            "actions": [
                {
                    "action_type": "ModifyProperties",
                    "order": 0,
                    "action_configuration": {
                        "classification": "BenignPositive",
                        "classification_comment": "comments 1",
                        "severity": "High",
                        "secret": "secret-value",
                    }
                },
                {
                    "action_type": "RunPlaybook",
                    "order": 1,
                    "logic_app_resource_id": "123333",
                    "tenant_id": "111111",
                },
                {
                    "action_type": "ModifyProperties",
                    "order": 2,
                    "action_configuration": {
                        "classification": "FalsePositive"
                    }
                }
            ]
        })

        # change the action_type will disable the access to previous discriminator, event though the data still in _data
        v.actions[2]['actionType'] = "RunPlaybook"
        with self.assertRaises(AAZUnknownFieldError):
            v.actions[2].action_configuration
        self.assertTrue(v.actions[2].logic_app_resource_id._is_patch)
        v.actions[2].logic_app_resource_id = "678"

        # will not serialize action_configuration
        self.assertTrue(v.to_serialized_data() == {
            "actions": [
                {
                    "action_type": "ModifyProperties",
                    "order": 0,
                    "action_configuration": {
                        "classification": "BenignPositive",
                        "classification_comment": "comments 1",
                        "severity": "High",
                        "secret": "secret-value",
                    }
                },
                {
                    "action_type": "RunPlaybook",
                    "order": 1,
                    "logic_app_resource_id": "123333",
                    "tenant_id": "111111",
                },
                {
                    "action_type": "RunPlaybook",
                    "order": 2,
                    "logic_app_resource_id": "678",
                }
            ]
        })

        # change back the action_type, action_configuration show again
        v.actions[2].action_type = "ModifyProperties"
        self.assertTrue(v.to_serialized_data() == {
            "actions": [
                {
                    "action_type": "ModifyProperties",
                    "order": 0,
                    "action_configuration": {
                        "classification": "BenignPositive",
                        "classification_comment": "comments 1",
                        "severity": "High",
                        "secret": "secret-value",
                    }
                },
                {
                    "action_type": "RunPlaybook",
                    "order": 1,
                    "logic_app_resource_id": "123333",
                    "tenant_id": "111111",
                },
                {
                    "action_type": "ModifyProperties",
                    "order": 2,
                    "action_configuration": {
                        "classification": "FalsePositive"
                    }
                }
            ]
        })

        # test client flatten and secret hidden
        self.assertTrue(AAZCommand.deserialize_output(v, client_flatten=True, secret_hidden=True) == {
            "actions": [
                {
                    "action_type": "ModifyProperties",
                    "order": 0,
                    "classification": "BenignPositive",
                    "classification_comment": "comments 1",
                    "severity": "High",
                },
                {
                    "action_type": "RunPlaybook",
                    "order": 1,
                    "logic_app_resource_id": "123333",
                    "tenant_id": "111111",
                },
                {
                    "action_type": "ModifyProperties",
                    "order": 2,
                    "classification": "FalsePositive"
                }
            ]
        })

        self.assertTrue(AAZCommand.deserialize_output(v, client_flatten=True, secret_hidden=False) == {
            "actions": [
                {
                    "action_type": "ModifyProperties",
                    "order": 0,
                    "classification": "BenignPositive",
                    "classification_comment": "comments 1",
                    "severity": "High",
                    "secret": "secret-value",
                },
                {
                    "action_type": "RunPlaybook",
                    "order": 1,
                    "logic_app_resource_id": "123333",
                    "tenant_id": "111111",
                },
                {
                    "action_type": "ModifyProperties",
                    "order": 2,
                    "classification": "FalsePositive"
                }
            ]
        })

        self.assertTrue(AAZCommand.deserialize_output(v, client_flatten=False, secret_hidden=True) == {
            "actions": [
                {
                    "action_type": "ModifyProperties",
                    "order": 0,
                    "action_configuration": {
                        "classification": "BenignPositive",
                        "classification_comment": "comments 1",
                        "severity": "High",
                    }
                },
                {
                    "action_type": "RunPlaybook",
                    "order": 1,
                    "logic_app_resource_id": "123333",
                    "tenant_id": "111111",
                },
                {
                    "action_type": "ModifyProperties",
                    "order": 2,
                    "action_configuration": {
                        "classification": "FalsePositive"
                    }
                }
            ]
        })

        self.assertTrue(AAZCommand.deserialize_output(v, client_flatten=False, secret_hidden=False) == {
            "actions": [
                {
                    "action_type": "ModifyProperties",
                    "order": 0,
                    "action_configuration": {
                        "classification": "BenignPositive",
                        "classification_comment": "comments 1",
                        "severity": "High",
                        "secret": "secret-value",
                    }
                },
                {
                    "action_type": "RunPlaybook",
                    "order": 1,
                    "logic_app_resource_id": "123333",
                    "tenant_id": "111111",
                },
                {
                    "action_type": "ModifyProperties",
                    "order": 2,
                    "action_configuration": {
                        "classification": "FalsePositive"
                    }
                }
            ]
        })

    def test_aaz_types_process_patch_data(self):
        from azure.cli.core.aaz._field_type import AAZObjectType, AAZDictType, AAZListType, \
            AAZIntType, AAZStrType, AAZBoolType, AAZFloatType
        from azure.cli.core.aaz._field_value import AAZObject, AAZValuePatch

        model_schema = AAZObjectType()
        model_schema.tags = AAZDictType()
        model_schema.tags.Element = AAZStrType()

        model_schema.properties = AAZObjectType()
        model_schema.properties.enabled = AAZBoolType()
        model_schema.properties.closed = AAZBoolType()

        model_schema.subnets = AAZListType()
        model_schema.subnets.Element = AAZObjectType()
        model_schema.subnets.Element.score = AAZFloatType()
        model_schema.subnets.Element.count = AAZIntType()

        v = AAZObject(schema=model_schema, data=AAZValuePatch.build(model_schema))
        v_copy = AAZObject(schema=model_schema, data=AAZValuePatch.build(model_schema))

        v.tags['a'] = "11"
        _ = v.tags['b']
        self.assertTrue(v.tags._is_patch and not v.tags['a']._is_patch)
        self.assertTrue(v.tags['b']._is_patch)

        data = model_schema.tags.process_data(v.tags)
        self.assertTrue(isinstance(data, AAZValuePatch))
        self.assertTrue(data.data['a'] == '11')
        self.assertTrue(isinstance(data.data['b'], AAZValuePatch))
        v_copy.tags = v.tags
        self.assertTrue(v_copy.tags._is_patch and not v_copy.tags['a']._is_patch)
        self.assertTrue(v_copy.tags['b']._is_patch)

        v.properties.enabled = False
        _ = v.properties.closed
        self.assertTrue(v.properties._is_patch and not v.properties.enabled._is_patch)
        self.assertTrue(v.properties.closed._is_patch)

        data = model_schema.properties.process_data(v.properties)
        self.assertTrue(isinstance(data, AAZValuePatch))
        self.assertTrue(data.data['enabled'] is False)
        self.assertTrue(isinstance(data.data['closed'], AAZValuePatch))
        v_copy.properties = v.properties
        self.assertTrue(v_copy.properties._is_patch and not v_copy.properties.enabled._is_patch)
        self.assertTrue(v_copy.properties.closed._is_patch)

        v.subnets[0].score = 1.1
        v.subnets[0].count = 1
        _ = v.subnets[1].score
        _ = v.subnets[1].count
        self.assertTrue(v.subnets[0]._is_patch and not v.subnets[0].score._is_patch and not v.subnets[0].count._is_patch)
        self.assertTrue(v.subnets[1]._is_patch and v.subnets[1].score._is_patch and v.subnets[1].count._is_patch)

        data = model_schema.subnets.process_data(v.subnets)
        self.assertTrue(isinstance(data, AAZValuePatch))
        self.assertTrue(isinstance(data.data[0], AAZValuePatch))
        self.assertTrue(data.data[0].data['score'] == 1.1)
        self.assertTrue(data.data[0].data['count'] == 1)
        self.assertTrue(isinstance(data.data[1], AAZValuePatch))
        self.assertTrue(isinstance(data.data[1].data['score'], AAZValuePatch))
        self.assertTrue(isinstance(data.data[1].data['count'], AAZValuePatch))

        v_copy.subnets = v.subnets

        self.assertTrue(v_copy.subnets[0]._is_patch and not v_copy.subnets[0].score._is_patch and not v_copy.subnets[0].count._is_patch)
        self.assertTrue(v_copy.subnets[1]._is_patch and v_copy.subnets[1].score._is_patch and v_copy.subnets[1].count._is_patch)

    def test_aaz_equal_comparison(self):
        from azure.cli.core.aaz._field_type import AAZObjectType, AAZListType, AAZFreeFormDictType, AAZDictType, AAZIntType, AAZStrType, AAZBoolType, AAZFloatType
        from azure.cli.core.aaz._field_value import AAZObject
        model_schema = AAZObjectType()

        model_schema.name = AAZStrType()

        model_schema.list = AAZListType()
        element = model_schema.list.Element = AAZObjectType()
        element.name = AAZStrType()

        model_schema.e_list = AAZListType()
        element = model_schema.e_list.Element = AAZStrType()

        model_schema.props = AAZObjectType()
        model_schema.props.name = AAZStrType()

        model_schema.e_props = AAZObjectType()
        model_schema.e_props.name = AAZStrType()

        model_schema.dict = AAZDictType()
        element = model_schema.dict.Element = AAZObjectType()
        element.name = AAZStrType()

        model_schema.e_dict = AAZDictType()
        element = model_schema.e_dict.Element = AAZStrType()

        model_schema.ff_dict = AAZFreeFormDictType()
        model_schema.e_ff_dict = AAZFreeFormDictType()

        data = {
            "name": "name",
            "list": [
                {
                    "name": "a"
                },
                {
                    "name": "b"
                }
            ],
            "props": {
                "name": "props"
            },
            "dict": {
                "key1": {
                    "name": "key1"
                },
                "key2": {
                    "name": "key2"
                }
            },
            "ff_dict": {
                "key1": {
                    "name": "key1"
                },
                "key2": {
                    "name": "key2"
                }
            },
        }
        v1 = AAZObject(model_schema, data=model_schema.process_data(data))
        v2 = AAZObject(model_schema, data=model_schema.process_data(data))
        self.assertEqual(v1, v2)
        self.assertEqual(v1, data)

        self.assertEqual(v1.name, v2.name)
        self.assertEqual(v1.list, v2.list)
        self.assertEqual(v1.props, v2.props)
        self.assertEqual(v1.dict, v2.dict)
        self.assertTrue(v1.ff_dict, v2.ff_dict)
        self.assertEqual(v1.ff_dict, v1.dict)

        self.assertEqual(v1.e_list, v2.e_list)
        self.assertEqual(v1.e_props, v2.e_props)
        self.assertEqual(v1.e_dict, v2.e_dict)
        self.assertEqual(v1.e_ff_dict, v2.e_ff_dict)
        self.assertEqual(v1.e_ff_dict, v1.e_dict)

        # not exist compare
        self.assertEqual(v1.list[10], v2.list[10])
        self.assertEqual(v1.dict["NotExist"], v2.dict["NotExist"])
